import { print } from 'graphql';
import { Anyware } from '../../lib/anyware/__.js';
import { parseExecutionResult } from '../../lib/graphqlHTTP.js';
import { CONTENT_TYPE_GQL, CONTENT_TYPE_JSON } from '../../lib/http.js';
import { casesExhausted } from '../../lib/prelude.js';
import { execute } from '../0_functions/execute.js';
import { SelectionSet } from '../3_SelectionSet/__.js';
import * as Result from '../4_ResultSet/customScalars.js';
const getRootIndexOrThrow = (context, rootTypeName) => {
    // @ts-expect-error
    // eslint-disable-next-line
    const rootIndex = context.schemaIndex.Root[rootTypeName];
    if (!rootIndex)
        throw new Error(`Root type not found: ${rootTypeName}`);
    return rootIndex;
};
export const hookNamesOrderedBySequence = [`encode`, `pack`, `exchange`, `unpack`, `decode`];
export const anyware = Anyware.create({
    hookNamesOrderedBySequence,
    hooks: {
        encode: {
            slots: {
                body: (input) => {
                    return JSON.stringify({
                        query: input.query,
                        variables: input.variables,
                        operationName: input.operationName,
                    });
                },
            },
            run: ({ input, slots }) => {
                let document;
                let variables = undefined;
                let operationName = undefined;
                switch (input.interface) {
                    case `raw`: {
                        const documentPrinted = typeof input.document === `string`
                            ? input.document
                            : print(input.document);
                        document = documentPrinted;
                        variables = input.variables;
                        operationName = input.operationName;
                        break;
                    }
                    case `typed`: {
                        // todo turn inputs into variables
                        variables = undefined;
                        document = SelectionSet.Print.rootTypeSelectionSet(input.context, getRootIndexOrThrow(input.context, input.rootTypeName), input.selection);
                        break;
                    }
                    default:
                        throw casesExhausted(input);
                }
                switch (input.transport) {
                    case `http`: {
                        const body = slots.body({
                            query: document,
                            variables,
                            operationName,
                        });
                        return {
                            ...input,
                            url: input.schema,
                            body,
                        };
                    }
                    case `memory`: {
                        return {
                            ...input,
                            schema: input.schema,
                            query: document,
                            variables,
                            operationName,
                        };
                    }
                }
            },
        },
        pack: ({ input }) => {
            switch (input.transport) {
                case `memory`: {
                    return input;
                }
                case `http`: {
                    const headers = new Headers(input.headers);
                    // @see https://graphql.github.io/graphql-over-http/draft/#sec-Accept
                    headers.set(`accept`, CONTENT_TYPE_GQL);
                    // @see https://graphql.github.io/graphql-over-http/draft/#sec-POST
                    // todo if body is something else, say upload extension turns it into a FormData, then fetch will automatically set the content-type header.
                    // ... however we should not rely on that behavior, and instead error here if there is no content type header and we cannot infer it here?
                    if (typeof input.body === `string`) {
                        headers.set(`content-type`, CONTENT_TYPE_JSON);
                    }
                    return {
                        ...input,
                        request: {
                            url: input.url,
                            body: input.body,
                            method: `POST`,
                            headers,
                        },
                    };
                }
                default:
                    throw casesExhausted(input);
            }
        },
        exchange: {
            slots: {
                fetch: (request) => {
                    return fetch(request);
                },
            },
            run: async ({ input, slots }) => {
                switch (input.transport) {
                    case `http`: {
                        const response = await slots.fetch(new Request(input.request.url, {
                            method: input.request.method,
                            headers: input.request.headers,
                            body: input.request.body,
                        }));
                        return {
                            ...input,
                            response,
                        };
                    }
                    case `memory`: {
                        const result = await execute({
                            schema: input.schema,
                            document: input.query,
                            variables: input.variables,
                            operationName: input.operationName,
                        });
                        return {
                            ...input,
                            result,
                        };
                    }
                    default:
                        throw casesExhausted(input);
                }
            },
        },
        unpack: async ({ input }) => {
            switch (input.transport) {
                case `http`: {
                    // todo 1 if response is missing header of content length then .json() hangs forever.
                    // todo 1 firstly consider a timeout, secondly, if response is malformed, then don't even run .json()
                    // todo 2 if response is e.g. 404 with no json body, then an error is thrown because json parse cannot work, not gracefully handled here
                    const json = await input.response.json();
                    const result = parseExecutionResult(json);
                    return {
                        ...input,
                        result,
                    };
                }
                case `memory`: {
                    return {
                        ...input,
                        result: input.result,
                    };
                }
                default:
                    throw casesExhausted(input);
            }
        },
        decode: ({ input }) => {
            switch (input.interface) {
                // todo this depends on the return mode
                case `raw`: {
                    switch (input.transport) {
                        case `http`: {
                            return {
                                ...input.result,
                                response: input.response,
                            };
                        }
                        case `memory`: {
                            return input.result;
                        }
                        default:
                            throw casesExhausted(input);
                    }
                }
                case `typed`: {
                    // todo optimize
                    // 1. Generate a map of possible custom scalar paths (tree structure)
                    // 2. When traversing the result, skip keys that are not in the map
                    const dataDecoded = Result.decode(getRootIndexOrThrow(input.context, input.rootTypeName), input.result.data);
                    // console.log(8, Object.keys({ ...input.result, data: dataDecoded }))
                    switch (input.transport) {
                        case `memory`: {
                            return { ...input.result, data: dataDecoded };
                        }
                        case `http`: {
                            return { ...input.result, data: dataDecoded, response: input.response };
                        }
                        default:
                            throw casesExhausted(input);
                    }
                }
                default:
                    throw casesExhausted(input);
            }
        },
    },
    // todo expose return handling as part of the pipeline?
    // would be nice but alone would not yield type safe return handling
    // still, while figuring the type story out, might be a useful escape hatch for some cases...
});
//# sourceMappingURL=core.js.map